from __future__ import annotations

from pathlib import Path
from typing import Tuple

from typing_extensions import Self
from upath import UPath

from hats.catalog.dataset.table_properties import TableProperties
from hats.io import file_io


class Dataset:
    """A base HATS dataset that contains a properties file
    and the data contained in parquet files"""

    def __init__(self, catalog_info: TableProperties, catalog_path: str | Path | UPath | None = None) -> None:
        """Initializes a Dataset

        Args:
            catalog_info: A TableProperties object with the catalog metadata
            catalog_path: If the catalog is stored on disk, specify the location of the catalog
                Does not load the catalog from this path, only store as metadata
        """
        self.catalog_info = catalog_info
        self.catalog_name = self.catalog_info.catalog_name

        self.catalog_path = catalog_path
        self.on_disk = catalog_path is not None
        self.catalog_base_dir = file_io.get_upath(self.catalog_path)

    @classmethod
    def read_hats(cls, catalog_path: str | Path | UPath) -> Self:
        """Reads a HATS Catalog from a HATS directory

        Args:
            catalog_path: path to the root directory of the catalog

        Returns:
            The initialized catalog object
        """
        catalog_base_dir = file_io.get_upath(catalog_path)
        cls._check_files_exist(catalog_base_dir)
        args = cls._read_args(catalog_base_dir)
        kwargs = cls._read_kwargs(catalog_base_dir)
        return cls(*args, **kwargs)

    @classmethod
    def _read_args(cls, catalog_base_dir: str | Path | UPath) -> Tuple[TableProperties]:
        catalog_info = TableProperties.read_from_dir(catalog_base_dir)
        return (catalog_info,)

    @classmethod
    def _read_kwargs(cls, catalog_base_dir: str | Path | UPath) -> dict:
        return {"catalog_path": catalog_base_dir}

    @classmethod
    def _check_files_exist(cls, catalog_base_dir: str | Path | UPath):
        if not file_io.does_file_or_directory_exist(catalog_base_dir):
            raise FileNotFoundError(f"No directory exists at {str(catalog_base_dir)}")
        if not file_io.does_file_or_directory_exist(catalog_base_dir / "properties"):
            raise FileNotFoundError(f"No properties file found where expected: {str(catalog_base_dir)}")



import pytest

from hats.pixel_math import HealpixPixel
from hats.pixel_tree.pixel_tree import PixelTree


@pytest.fixture
def pixel_trees_dir(test_data_dir):
    return test_data_dir / "pixel_trees"


@pytest.fixture
def pixel_tree_1():
    return PixelTree.from_healpix([HealpixPixel(0, 11)])


@pytest.fixture
def pixel_tree_2(pixel_list_breadth_first):
    return PixelTree.from_healpix(pixel_list_breadth_first)


@pytest.fixture
def pixel_tree_3():
    return PixelTree.from_healpix(
        [
            HealpixPixel(0, 8),
            HealpixPixel(1, 36),
            HealpixPixel(1, 37),
            HealpixPixel(1, 40),
            HealpixPixel(1, 42),
            HealpixPixel(1, 43),
            HealpixPixel(1, 44),
            HealpixPixel(1, 46),
            HealpixPixel(1, 47),
        ]
    )


@pytest.fixture
def aligned_trees_2_3_inner():
    return PixelTree.from_healpix(
        [
            HealpixPixel(1, 33),
            HealpixPixel(1, 35),
            HealpixPixel(1, 40),
            HealpixPixel(1, 42),
            HealpixPixel(1, 43),
            HealpixPixel(1, 44),
            HealpixPixel(1, 46),
            HealpixPixel(2, 128),
            HealpixPixel(2, 130),
            HealpixPixel(2, 131),
        ]
    )


@pytest.fixture
def aligned_trees_2_3_left():
    return PixelTree.from_healpix(
        [
            HealpixPixel(1, 33),
            HealpixPixel(1, 35),
            HealpixPixel(1, 40),
            HealpixPixel(1, 41),
            HealpixPixel(1, 42),
            HealpixPixel(1, 43),
            HealpixPixel(1, 44),
            HealpixPixel(1, 45),
            HealpixPixel(1, 46),
            HealpixPixel(2, 128),
            HealpixPixel(2, 130),
            HealpixPixel(2, 131),
        ]
    )


@pytest.fixture
def aligned_trees_2_3_right():
    return PixelTree.from_healpix(
        [
            HealpixPixel(1, 33),
            HealpixPixel(1, 34),
            HealpixPixel(1, 35),
            HealpixPixel(1, 36),
            HealpixPixel(1, 37),
            HealpixPixel(1, 40),
            HealpixPixel(1, 42),
            HealpixPixel(1, 43),
            HealpixPixel(1, 44),
            HealpixPixel(1, 46),
            HealpixPixel(1, 47),
            HealpixPixel(2, 128),
            HealpixPixel(2, 129),
            HealpixPixel(2, 130),
            HealpixPixel(2, 131),
        ]
    )


@pytest.fixture
def aligned_trees_2_3_outer():
    return PixelTree.from_healpix(
        [
            HealpixPixel(1, 33),
            HealpixPixel(1, 34),
            HealpixPixel(1, 35),
            HealpixPixel(1, 36),
            HealpixPixel(1, 37),
            HealpixPixel(1, 40),
            HealpixPixel(1, 41),
            HealpixPixel(1, 42),
            HealpixPixel(1, 43),
            HealpixPixel(1, 44),
            HealpixPixel(1, 45),
            HealpixPixel(1, 46),
            HealpixPixel(1, 47),
            HealpixPixel(2, 128),
            HealpixPixel(2, 129),
            HealpixPixel(2, 130),
            HealpixPixel(2, 131),
        ]
    )



import pytest

from hats.pixel_math import HealpixPixel
from hats.pixel_math.healpix_pixel_convertor import get_healpix_pixel, get_healpix_tuple


def test_get_healpix_pixel_can_take_either_argument():
    order = 5
    pixel = 10
    for arg in [HealpixPixel(order, pixel), (order, pixel)]:
        returned_pixel = get_healpix_pixel(arg)
        assert isinstance(returned_pixel, HealpixPixel)
        assert returned_pixel.order == order
        assert returned_pixel.pixel == pixel


def test_get_healpix_pixel_raises_error_for_wrong_type():
    with pytest.raises(TypeError):
        get_healpix_pixel("test")


def test_get_healpix_pixel_raises_error_for_wrong_size_tuple():
    with pytest.raises(ValueError):
        get_healpix_pixel((1, 2, 3))


def test_get_healpix_tuple_can_take_either():
    order = 5
    pixel = 10
    for arg in [HealpixPixel(order, pixel), (order, pixel)]:
        returned_pixel = get_healpix_tuple(arg)
        assert isinstance(returned_pixel, tuple)
        assert returned_pixel[0] == order
        assert returned_pixel[1] == pixel


def test_get_healpix_tuple_raises_error_for_wrong_type():
    with pytest.raises(TypeError):
        get_healpix_tuple("test")


def test_get_healpix_tuple_raises_error_for_wrong_size_tuple():
    with pytest.raises(ValueError):
        get_healpix_tuple((1, 2, 3))



from __future__ import annotations

from collections.abc import Iterable
from typing import List, Tuple

import numpy as np
from mocpy import MOC

import hats.pixel_math.healpix_shim as hp
from hats.pixel_math.polygon_filter import SphericalCoordinates


def generate_box_moc(ra: Tuple[float, float] | None, dec: Tuple[float, float] | None, order: int) -> MOC:
    """Generates a MOC object that covers the specified box area

    Args:
        ra (Tuple[float, float]): Right ascension range, in [0,360] degrees
        dec (Tuple[float, float]): Declination range, in [-90,90] degrees
        order (int): Maximum order of the moc to generate the box at

    Returns:
        a MOC object that covers the specified box
    """
    filter_moc = None
    ra_search_moc, dec_search_moc = None, None

    if ra is not None:
        ra_search_moc = _generate_ra_strip_moc(ra, order)
        filter_moc = ra_search_moc
    if dec is not None:
        dec_search_moc = _generate_dec_strip_moc(dec, order)
        filter_moc = dec_search_moc
    if ra_search_moc is not None and dec_search_moc is not None:
        filter_moc = ra_search_moc.intersection(dec_search_moc)
    return filter_moc


def wrap_ra_angles(ra: np.ndarray | Iterable | int | float) -> np.ndarray:
    """Wraps angles to the [0,360] degree range.

    Args:
        ra (ndarray | Iterable | int | float): Right ascension values, in degrees

    Returns:
        A numpy array of right ascension values, wrapped to the [0,360] degree range.
    """
    return np.asarray(ra, dtype=float) % 360


def _generate_ra_strip_moc(ra_range: Tuple[float, float], order: int) -> MOC:
    """Generates a pixel_tree filled with leaf nodes at a given order that overlap with the ra region."""
    # Subdivide polygons (if needed) in two smaller polygons of at most 180 degrees
    division_ra = _get_division_ra(ra_range)

    if division_ra is not None:
        pixels_in_range = _get_pixels_for_subpolygons(
            [
                # North Pole
                [(0, 90), (ra_range[0], 0), (division_ra, 0)],
                [(0, 90), (division_ra, 0), (ra_range[1], 0)],
                # South Pole
                [(ra_range[0], 0), (0, -90), (division_ra, 0)],
                [(division_ra, 0), (0, -90), (ra_range[1], 0)],
            ],
            order,
        )
    else:
        # We assume the polygon of smallest area by default
        # e.g. for ra_ranges of [10,50], [200,10] or [350,10]
        pixels_in_range = _get_pixels_for_subpolygons(
            [
                # North Pole
                [(0, 90), (ra_range[0], 0), (ra_range[1], 0)],
                # South Pole
                [(ra_range[0], 0), (0, -90), (ra_range[1], 0)],
            ],
            order,
        )

    orders = np.full(pixels_in_range.shape, fill_value=order)
    return MOC.from_healpix_cells(ipix=pixels_in_range, depth=orders, max_depth=order)


def _generate_dec_strip_moc(dec_range: Tuple[float, float], order: int) -> MOC:
    """Generates a pixel_tree filled with leaf nodes at a given order that overlap with the dec region."""
    nside = hp.order2nside(order)
    # Convert declination values to colatitudes, in radians, and revert their order
    colat_rad = np.sort(np.radians([90 - dec if dec > 0 else 90 + abs(dec) for dec in dec_range]))
    # Figure out which pixels in nested order are contained in the declination band
    pixels_in_range = hp.ring2nest(
        nside, hp.query_strip(nside, theta1=colat_rad[0], theta2=colat_rad[1], inclusive=True)
    )
    orders = np.full(pixels_in_range.shape, fill_value=order)
    return MOC.from_healpix_cells(ipix=pixels_in_range, depth=orders, max_depth=order)


def _get_division_ra(ra_range: Tuple[float, float]) -> float | None:
    """Gets the division angle for the right ascension region. This is crucial for the edge
    cases where we need to split up polygons wider than 180 degrees into smaller polygons."""
    division_ra = None
    if ra_range[0] > ra_range[1] and 360 - ra_range[0] + ra_range[1] >= 180:
        # e.g. edge case of [350, 170] or [180, 0], we want the wider polygon
        division_ra = (ra_range[1] - 360 + ra_range[0]) / 2
    elif ra_range[0] < ra_range[1] and ra_range[1] - ra_range[0] >= 180:
        # e.g. edge case of [10, 200] or [0, 180], we want the wider polygon
        division_ra = ra_range[0] + (ra_range[1] - ra_range[0]) / 2
    return division_ra


def _get_pixels_for_subpolygons(polygons: List[List[SphericalCoordinates]], order: int) -> np.ndarray:
    """Gets the unique pixels for a set of sub-polygons."""
    nside = hp.order2nside(order)
    all_polygon_pixels = []
    for vertices in polygons:
        vertices = hp.ang2vec(*np.array(vertices).T, lonlat=True)
        pixels = hp.query_polygon(nside, vertices, inclusive=True, nest=True)
        all_polygon_pixels.append(pixels)
    return np.unique(np.concatenate(all_polygon_pixels, 0))



"""Tests of margin bounding utility functions"""

import numpy as np
import numpy.testing as npt
import pytest

import hats.pixel_math as pm


def test_check_margin_bounds():
    """Make sure check_margin_bounds works properly"""

    ras = np.array([42.4704538, 56.25, 56.25, 50.61225197])
    decs = np.array([1.4593925, 9.6, 10.0, 14.4767556])

    checks = pm.check_margin_bounds(ras, decs, 3, 4, 360.0)

    expected = np.array([False, True, False, True])

    npt.assert_array_equal(checks, expected)


def test_check_margin_bounds_low_order():
    """Make sure check_margin_bounds works when using a low order healpixel"""

    ras = np.array(
        [
            142.4704538,
            45.09,
            45.2,
            37.31343956517391,
            42.649354753311535,
            32.62796809350278,
            39.89468227954832,
            27.718121934039974,
        ]
    )

    decs = np.array(
        [
            1.4593925,
            0.0,
            0.0,
            6.566326903165274,
            2.005185097251452,
            10.597884275167646,
            4.465967883812584,
            14.959672304191956,
        ]
    )

    checks = pm.check_margin_bounds(ras, decs, 0, 4, 360.0)

    expected = np.array([False, True, False, True, True, True, True, True])

    npt.assert_array_equal(checks, expected)


def test_check_polar_margin_bounds_north():
    """Make sure check_polar_margin_bounds works at the north pole"""
    order = 0
    pix = 1

    ras = np.array([89, -179, -45])
    decs = np.array([89.9, 89.9, 85.0])

    vals = pm.check_margin_bounds(ras, decs, order, pix, 360.0)

    expected = np.array([True, True, False])

    npt.assert_array_equal(vals, expected)


def test_check_polar_margin_bounds_south():
    """Make sure check_polar_margin_bounds works at the south pole"""
    order = 0
    pix = 9

    ras = np.array([89, -179, -45])
    decs = np.array([-89.9, -89.9, -85.0])

    vals = pm.check_margin_bounds(ras, decs, order, pix, 360.0)

    expected = np.array([True, True, False])

    npt.assert_array_equal(vals, expected)


def test_check_margin_bounds_uneven():
    """Ensure check_margin_bounds fails when ra and dec arrays are unbalanced."""

    ras = np.array([42.4704538, 56.25, 56.25, 50.61225197])
    decs = np.array([1.4593925, 9.6, 10.0])

    with pytest.raises(ValueError, match="length of r_asc"):
        pm.check_margin_bounds(ras, decs, 3, 4, 360.0)


def test_check_margin_bounds_empty():
    """Ensure check_margin_bounds works when passed empty coordinate arrays."""

    ras = np.array([])
    decs = np.array([])

    checks = pm.check_margin_bounds(ras, decs, 3, 4, 360.0)

    expected = np.array([])

    npt.assert_array_equal(checks, expected)



"""Tests of histogram calculations"""

import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest

import hats.pixel_math as hist
import hats.pixel_math.healpix_shim as hp


def test_small_sky_same_pixel():
    """Test partitioning two objects into the same large bucket"""

    data = pd.DataFrame(
        data=[[700, 282.5, -58.5], [701, 299.5, -48.5]],
        columns=["id", "ra", "dec"],
    )

    result = hist.generate_histogram(
        data=data,
        highest_order=0,
    )

    assert len(result) == 12

    expected = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2]
    npt.assert_array_equal(result, expected)
    assert (result == expected).all()


def test_column_names_error():
    """Test with non-default column names (without specifying column names)"""

    data = pd.DataFrame(
        data=[[700, 282.5, -58.5], [701, 299.5, -48.5]],
        columns=["id", "ra_mean", "dec_mean"],
    )

    with pytest.raises(ValueError, match="Invalid column names"):
        hist.generate_histogram(
            data=data,
            highest_order=0,
        )


def test_column_names():
    """Test with non-default column names"""
    data = pd.DataFrame(
        data=[[700, 282.5, -58.5], [701, 299.5, -48.5]],
        columns=["id", "ra_mean", "dec_mean"],
    )

    result = hist.generate_histogram(
        data=data,
        highest_order=0,
        ra_column="ra_mean",
        dec_column="dec_mean",
    )

    assert len(result) == 12

    expected = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2]
    npt.assert_array_equal(result, expected)
    assert (result == expected).all()


def test_alignment_wrong_size():
    """Check that the method raises error when the input histogram is not the expected size."""
    initial_histogram = np.asarray([0, 0, 0, 0, 0, 0, 0, 0, 0, 131])
    with pytest.raises(ValueError, match="histogram is not the right size"):
        hist.generate_alignment(initial_histogram, highest_order=0, threshold=250)


def test_alignment_exceeds_threshold_order0():
    """Check that the method raises error when some pixel exceeds the threshold."""
    initial_histogram = np.asarray([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 131])
    with pytest.raises(ValueError, match="exceeds threshold"):
        hist.generate_alignment(initial_histogram, highest_order=0, threshold=20)


def test_alignment_lowest_order_too_large():
    """Check that the method raises error when some pixel exceeds the threshold."""
    initial_histogram = hist.empty_histogram(1)
    with pytest.raises(ValueError, match="lowest_order"):
        hist.generate_alignment(initial_histogram, highest_order=1, lowest_order=2, threshold=20)


def test_alignment_exceeds_threshold_order2():
    """Check that the method raises error when some pixel exceeds the threshold."""
    initial_histogram = hist.empty_histogram(2)
    filled_pixels = [4, 11, 14, 13, 5, 7, 8, 9, 11, 23, 4, 4, 17, 0, 1, 0]
    initial_histogram[176:] = filled_pixels[:]
    with pytest.raises(ValueError, match="exceeds threshold"):
        hist.generate_alignment(initial_histogram, highest_order=2, threshold=20)


@pytest.mark.parametrize("drop_empty_siblings", [True, False])
def test_alignment_small_sky_order0(drop_empty_siblings):
    """Create alignment from small sky's distribution at order 0"""
    initial_histogram = np.asarray([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 131])
    result = hist.generate_alignment(
        initial_histogram, highest_order=0, threshold=250, drop_empty_siblings=drop_empty_siblings
    )

    expected = np.full(12, None)
    expected[11] = (0, 11, 131)

    npt.assert_array_equal(result, expected)


@pytest.mark.parametrize("drop_empty_siblings", [True, False])
def test_alignment_small_sky_order1(drop_empty_siblings):
    """Create alignment from small sky's distribution at order 1"""
    initial_histogram = hist.empty_histogram(1)
    filled_pixels = [42, 29, 42, 18]
    initial_histogram[44:] = filled_pixels[:]
    result = hist.generate_alignment(
        initial_histogram, highest_order=1, threshold=250, drop_empty_siblings=drop_empty_siblings
    )

    expected = np.full(48, None)
    expected[44:] = [(0, 11, 131), (0, 11, 131), (0, 11, 131), (0, 11, 131)]

    npt.assert_array_equal(result, expected)


def test_alignment_small_sky_order1_empty_siblings():
    """Create alignment from small sky's distribution at order 1"""
    initial_histogram = hist.empty_histogram(1)
    initial_histogram[44] = 100
    result = hist.generate_alignment(
        initial_histogram, highest_order=1, threshold=250, drop_empty_siblings=True
    )

    expected = np.full(48, None)
    expected[44] = (1, 44, 100)

    npt.assert_array_equal(result, expected)


@pytest.mark.parametrize("drop_empty_siblings", [True, False])
def test_alignment_small_sky_order2(drop_empty_siblings):
    """Create alignment from small sky's distribution at order 2"""
    initial_histogram = hist.empty_histogram(2)
    filled_pixels = [4, 11, 14, 13, 5, 7, 8, 9, 11, 23, 4, 4, 17, 0, 1, 0]
    initial_histogram[176:] = filled_pixels[:]
    result = hist.generate_alignment(
        initial_histogram, highest_order=2, threshold=250, drop_empty_siblings=drop_empty_siblings
    )

    expected = np.full(hp.order2npix(2), None)
    tuples = [
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
        (0, 11, 131),
    ]
    expected[176:192] = tuples

    npt.assert_array_equal(result, expected)


@pytest.mark.parametrize("drop_empty_siblings", [True, False])
@pytest.mark.timeout(20)
def test_alignment_even_sky(drop_empty_siblings):
    """Create alignment from an even distribution at order 7"""
    initial_histogram = np.full(hp.order2npix(7), 40)
    result = hist.generate_alignment(
        initial_histogram, highest_order=7, threshold=1_000, drop_empty_siblings=drop_empty_siblings
    )
    # everything maps to order 5, given the density
    for mapping in result:
        assert mapping[0] == 5

    result = hist.generate_alignment(
        initial_histogram,
        highest_order=7,
        lowest_order=7,
        threshold=1_000,
        drop_empty_siblings=drop_empty_siblings,
    )
    # everything maps to order 7 (would be 5, but lowest of 7 is enforced)
    for mapping in result:
        assert mapping[0] == 7



"""Set of convenience methods for testing file contents"""

import re

import numpy.testing as npt
import pyarrow.parquet as pq
import pytest

from hats.io import file_io
from hats.io.file_io.file_io import load_text_file
from hats.io.file_io.file_pointer import does_file_or_directory_exist

# pylint: disable=missing-function-docstring, redefined-outer-name


@pytest.fixture
def assert_text_file_matches():
    def assert_text_file_matches(expected_lines, file_name):
        """Convenience method to read a text file and compare the contents, line for line.

        When file contents get even a little bit big, it can be difficult to see
        the difference between an actual file and the expected contents without
        increased testing verbosity. This helper compares files line-by-line,
        using the provided strings or regular expressions.

        Notes:
            Because we check strings as regular expressions, you may need to escape some
            contents of `expected_lines`.

        Args:
            expected_lines(:obj:`string array`) list of strings, formatted as regular expressions.
            file_name (str): fully-specified path of the file to read
        """
        assert does_file_or_directory_exist(file_name), f"file not found [{file_name}]"
        contents = load_text_file(file_name)

        assert len(expected_lines) == len(
            contents
        ), f"files not the same length ({len(contents)} vs {len(expected_lines)})"
        for i, expected in enumerate(expected_lines):
            assert re.match(expected, contents[i]), (
                f"files do not match at line {i+1} " f"(actual: [{contents[i]}] vs expected: [{expected}])"
            )

    return assert_text_file_matches


@pytest.fixture
def check_parquet_schema():
    def check_parquet_schema(file_name, expected_schema, expected_num_row_groups=1):
        """Check parquet schema against expectations"""
        assert file_io.does_file_or_directory_exist(file_name), f"file not found [{file_name}]"

        single_metadata = file_io.read_parquet_metadata(file_name)
        schema = single_metadata.schema.to_arrow_schema()

        assert len(schema) == len(
            expected_schema
        ), f"object list not the same size ({len(schema)} vs {len(expected_schema)})"

        npt.assert_array_equal(schema.names, expected_schema.names)

        assert schema.equals(expected_schema, check_metadata=False)

        parquet_file = pq.ParquetFile(file_name)
        assert parquet_file.metadata.num_row_groups == expected_num_row_groups

        for row_index in range(0, parquet_file.metadata.num_row_groups):
            row_md = parquet_file.metadata.row_group(row_index)
            for column_index in range(0, row_md.num_columns):
                column_metadata = row_md.column(column_index)
                assert column_metadata.file_path.endswith(".parquet")

    return check_parquet_schema



"""Utilities for reading and writing catalog files"""

from .parquet_metadata import (
    read_row_group_fragments,
    row_group_stat_single_value,
    write_parquet_metadata,
    write_parquet_metadata_for_batches,
)
from .paths import (
    get_common_metadata_pointer,
    get_parquet_metadata_pointer,
    get_partition_info_pointer,
    get_point_map_file_pointer,
    pixel_catalog_file,
    pixel_directory,
)
